package lab.game;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;

import javafx.scene.canvas.GraphicsContext;
import javafx.scene.paint.Color;
import lab.Config;
import lombok.extern.log4j.Log4j2;

@Log4j2
public class Level {

	private Object entitiesLock = new Object();

	private boolean viewMode;

	private double width;
	private double height;
	private List<DrawableSimulable> entities;
	private List<DrawableSimulable> entitiesToAdd = new ArrayList<>();
	private List<DrawableSimulable> entitiesToRemove = new ArrayList<>();
	private Player player;
	private List<MonsterDeadLog> deadLogs = new LinkedList<>();

	public Level(double width, double height, int monsterCount, boolean viewMode) {
		this.width = width;
		this.height = height;
		this.viewMode = viewMode;
		player = new Player(this, new MyPoint(20, 250), new MyPoint(100, -20));
		entities = new ArrayList<>();
		entities.add(new NicerObstacle(this, new MyPoint(20, 150)));
		entities.add(new Obstacle(this, new MyPoint(300, 200), new MyDimension(80, 40)));
		entities.add(new Obstacle(this));
		entities.add(player);
		for (int i = 0; i < monsterCount * Config.getInstance().getMonsterMultiplier(); i++) {
			Monster monster = new Monster(this);
			monster.addDeadListener(m -> deadLogs.add(new MonsterDeadLog(LocalDateTime.now(), m.getPosition())));
			entities.add(monster);
		}
		initServerOrViewMode();
	}

	public void initServerOrViewMode() {
		if (!viewMode) {
			startServer();
		} else {
			connectToServer();
		}
	}

	public void startServer() {
		new Thread(() -> {
			try (ServerSocket listenSocket = new ServerSocket(4600)) {
				while (!Thread.currentThread().isInterrupted()) {
					try {
						Socket client = listenSocket.accept();
						new Thread(() -> handleViewer(client), "Viewer handler").start();
					} catch (IOException e) {
						e.printStackTrace();
					}
				}
			} catch (IOException e) {
				e.printStackTrace();
			}

		}).start();
	}

	public void handleViewer(Socket client) {
		try (OutputStream outputStream = client.getOutputStream();
				ObjectOutputStream objectOutputStream = new ObjectOutputStream(outputStream)) {
			while (!Thread.currentThread().isInterrupted()) {
				synchronized (entitiesLock) {
					objectOutputStream.reset();
					synchronized (entitiesLock) {
						objectOutputStream.writeObject(entities);
					}
					objectOutputStream.flush();
				}
				sleepForWhile();
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void sleepForWhile() {
		try {
			Thread.sleep(20);
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
		}
	}

	public void connectToServer() {
		new Thread(() -> {
			try (Socket socketToServer = new Socket("localhost", 4600);
					InputStream inputStream = socketToServer.getInputStream();
					ObjectInputStream objectInputStream = new ObjectInputStream(inputStream)) {
				while (!Thread.currentThread().isInterrupted()) {
					Object o = objectInputStream.readObject();
					if (o instanceof List l) {
						synchronized (entitiesLock) {
							entities = l;
						}
					}
				}
			} catch (IOException | ClassNotFoundException e) {
				e.printStackTrace();
			}

		}).start();
	}

	record MonsterDeadLog(LocalDateTime time, MyPoint position) {
	}

	public void printDeadLog() {
		for (MonsterDeadLog monsterDeadLog : deadLogs) {
			log.info(monsterDeadLog);
		}
	}

	public void draw(GraphicsContext gc) {
		gc.setFill(Color.WHITE);
		gc.clearRect(0, 0, width, height);
		synchronized (entitiesLock) {
			for (DrawableSimulable entity : entities) {
				entity.draw(gc);
			}
		}
	}

	public void simulate(double delay) {
		if (viewMode) {
			return;
		}
		synchronized (entitiesLock) {
			for (DrawableSimulable entity : entities) {
				entity.simulate(delay);
			}
			detectCollisions();
			entities.removeAll(entitiesToRemove);
			entities.addAll(entitiesToAdd);
			entitiesToAdd.clear();
			entitiesToRemove.clear();
			entities.sort(Comparator.comparing(DrawableSimulable::getZIndex));
		}
	}

	private void detectCollisions() {
		for (int i = 0; i < entities.size(); i++) {
			if (entities.get(i) instanceof Collisionable c1) {
				for (int j = i + 1; j < entities.size(); j++) {
					if (entities.get(j) instanceof Collisionable c2 && c1.intersect(c2.getBoundingBox())) {
						c1.hitBy(c2);
						c2.hitBy(c1);
					}
				}
			}
		}
	}

	public double getWidth() {
		return width;
	}

	public double getHeight() {
		return height;
	}

	public Player getPlayer() {
		return player;
	}

	public void add(DrawableSimulable entity) {
		if (viewMode) {
			return;
		}
		entitiesToAdd.add(entity);
	}

	public void remove(DrawableSimulable entity) {
		if (viewMode) {
			return;
		}
		entitiesToRemove.add(entity);
	}

}
